#! /usr/bin/python3

from pyln.spec.bolt7 import (channel_announcement, channel_update,
                             node_announcement)
from pyln.proto import ShortChannelId, PublicKey
from typing import Any, Dict, List, Set, Optional, Union

import io
import base64
import socket
import struct
import time

# These duplicate constants in lightning/common/gossip_store.h
GOSSIP_STORE_MAJOR_VERSION = (0 << 5)
GOSSIP_STORE_MAJOR_VERSION_MASK = 0xE0
GOSSIP_STORE_LEN_DELETED_BIT = 0x8000
GOSSIP_STORE_LEN_PUSH_BIT = 0x4000
GOSSIP_STORE_LEN_COMPLETE_BIT = 0x2000

# These duplicate constants in lightning/gossipd/gossip_store_wiregen.h
WIRE_GOSSIP_STORE_PRIVATE_CHANNEL = 4104
WIRE_GOSSIP_STORE_PRIVATE_UPDATE = 4102
WIRE_GOSSIP_STORE_DELETE_CHAN = 4103
WIRE_GOSSIP_STORE_ENDED = 4105
WIRE_GOSSIP_STORE_CHANNEL_AMOUNT = 4101


class LnFeatureBits(object):
    """ feature flags taken from bolts.git/09-features.md

 Flags are numbered from the least-significant bit, at bit 0 (i.e. 0x1,
 an _even_ bit). They are generally assigned in pairs so that features
 can be introduced as optional (_odd_ bits) and later upgraded to be compulsory
 (_even_ bits), which will be refused by outdated nodes:

 CONTEXT:
 * `I`: presented in the `init` message.
 * `N`: presented in the `node_announcement` messages
 * `C`: presented in the `channel_announcement` message.
 * `C-`: presented in the `channel_announcement` message, but always odd (optional).
 * `C+`: presented in the `channel_announcement` message, but always even (required).
 * `9`: presented in [BOLT 11](11-payment-encoding.md) invoices.

    FEATURE_NAME                                    # CONTEXT   # PRs
    ----------------------------------------------------------------- """
    OPTION_DATA_LOSS_PROTECT = 0                    # IN
    INITIAL_ROUTING_SYNC = 2                        # I
    OPTION_UPFRONT_SHUTDOWN_SCRIPT = 4              # IN
    GOSSIP_QUERIES = 6                              # IN
    VAR_ONION_OPTIN = 8                             # IN9
    GOSSIP_QUERIES_EX = 10                          # IN
    OPTION_STATIC_REMOTEKEY = 12                    # IN
    PAYMENT_SECRET = 14                             # IN9
    BASIC_MPP = 16                                  # IN9
    OPTION_SUPPORT_LARGE_CHANNEL = 18               # IN
    OPTION_ANCHOR_OUTPUTS = 20                      # IN
    OPTION_ANCHORS_ZERO_FEE_HTLC_TX = 22            # IN
    OPTION_SHUTDOWN_ANYSEGWIT = 26                  # IN
    OPTION_CHANNEL_TYPE = 44                        # IN
    OPTION_SCID_ALIAS = 46                          # IN
    OPTION_PAYMENT_METADATA = 48                    # 9
    OPTION_ZEROCONF = 50                            # IN

    OPTION_PROPOSED_ROUTE_BLINDING = 24             # IN9       #765 #798
    OPTION_PROPOSED_DUAL_FUND = 28                  # IN        #851 #1009
    OPTION_PROPOSED_ALTERNATIVE_FEERATES = 32       # IN        #1036
    OPTION_PROPOSED_QUIESCE = 34                    # IN        #869 #868
    OPTION_PROPOSED_ONION_MESSAGES = 38             # IN        #759
    OPTION_PROPOSED_WANT_PEER_BACKUP_STORAGE = 40   # IN        #881
    OPTION_PROPOSED_PROVIDE_PEER_BACKUP = 42        # IN        #881
    OPTION_PROPOSED_TRAMPOLINE_ROUTING = 56         # IN9       #836
    OPTION_PROPOSED_UPFRONT_FEE = 56                # IN9       #1052
    OPTION_PROPOSED_CLOSING_REJECTED = 60           # IN        #1016
    OPTION_PROPOSED_SPLICE = 62                     # IN        #863
    OPTION_PROPOSED_EXPERIMENTAL_SPLICE = 162       # IN        #863


def _parse_features(featurebytes):
    # featurebytes e.g.: [136, 160, 0, 8, 2, 105, 162]
    result = 0
    for byte in featurebytes:
        result <<= 8
        result |= byte
    return result


class GossipStoreMsgHeader(object):
    def __init__(self, buf: bytes, off: int):
        self.flags, self.length, self.crc, self.timestamp = struct.unpack('>HHII', buf)
        self.off = off
        self.deleted = (self.flags & GOSSIP_STORE_LEN_DELETED_BIT) != 0


class GossmapHalfchannel(object):
    """One direction of a GossmapChannel."""
    def __init__(self, channel: 'GossmapChannel', direction: int,
                 fields: Dict[str, Any], hdr: GossipStoreMsgHeader):
        assert direction in [0, 1], "direction can only be 0 or 1"
        self.channel = channel
        self.direction = direction
        self.source = channel.node1 if direction == 0 else channel.node2
        self.destination = channel.node2 if direction == 0 else channel.node1
        self.fields: Dict[str, Any] = fields
        self.hdr: GossipStoreMsgHeader = hdr

        self.timestamp: int = fields['timestamp']
        self.cltv_expiry_delta: int = fields['cltv_expiry_delta']
        self.htlc_minimum_msat: int = fields['htlc_minimum_msat']
        self.htlc_maximum_msat: Optional[int] = fields.get('htlc_maximum_msat', None)
        self.fee_base_msat: int = fields['fee_base_msat']
        self.fee_proportional_millionths: int = fields['fee_proportional_millionths']
        self.disabled = fields['channel_flags'] & 2 > 0

        # Cache the _scidd and hash to have faster operation later
        # Unfortunately the @final decorator only comes for python3.8
        self._scidd = f"{self.channel.scid}/{self.direction}"
        self._numscidd = direction << 63 | self.channel.scid.to_int()

    def __repr__(self):
        return f"GossmapHalfchannel[{self._scidd}]"

    def __eq__(self, other):
        if not isinstance(other, GossmapHalfchannel):
            return False
        return self._numscidd == other._numscidd

    def __str__(self):
        return self._scidd

    def __hash__(self):
        return self._numscidd


class GossmapNodeId(object):
    def __init__(self, buf: Union[bytes, str]):
        if isinstance(buf, str):
            buf = bytes.fromhex(buf)
        if len(buf) != 33 or (buf[0] != 2 and buf[0] != 3):
            raise ValueError("{} is not a valid node_id".format(buf.hex()))
        self.nodeid = buf

        self._hash = self.nodeid.__hash__()
        self._str = self.nodeid.hex()

    def to_pubkey(self) -> PublicKey:
        return PublicKey(self.nodeid)

    def __eq__(self, other):
        if not isinstance(other, GossmapNodeId):
            return False
        return self.nodeid.__eq__(other.nodeid)

    def __lt__(self, other):
        if not isinstance(other, GossmapNodeId):
            raise TypeError(f"Cannot compare GossmapNodeId with {type(other)}")
        return self.nodeid.__lt__(other.nodeid)  # yes, that works

    def __hash__(self):
        return self._hash

    def __repr__(self):
        return "GossmapNodeId[{}]".format(self.nodeid.hex())

    def __str__(self):
        return self._str

    @classmethod
    def from_str(cls, s: str):
        if s.startswith('0x'):
            s = s[2:]
        if len(s) != 66:
            raise ValueError(f"{s} is not a valid hexstring of a node_id")
        return cls(bytes.fromhex(s))


class GossmapChannel(object):
    """A channel: fields of channel_announcement are in .fields,
       optional updates are in .half_channels[0/1].fields """
    def __init__(self,
                 fields: Dict[str, Any],
                 scid: Union[ShortChannelId, str],
                 node1: 'GossmapNode',
                 node2: 'GossmapNode',
                 is_private: bool,
                 hdr: GossipStoreMsgHeader):
        self.fields: Dict[str, Any] = fields
        self.hdr: GossipStoreMsgHeader = hdr

        self.is_private = is_private
        self.scid = ShortChannelId.from_str(scid) if isinstance(scid, str) else scid
        self.node1 = node1
        self.node2 = node2
        self.satoshis = None
        self.half_channels: List[Optional[GossmapHalfchannel]] = [None, None]
        self.features = _parse_features(fields['features'])

    def _update_channel(self,
                        direction: int,
                        fields: Dict[str, Any],
                        hdr: GossipStoreMsgHeader):

        half = GossmapHalfchannel(self, direction, fields, hdr)
        self.half_channels[direction] = half

    def get_direction(self, direction: int):
        """ returns the GossmapHalfchannel if known by channel_update """
        assert direction in [0, 1], "direction can only be 0 or 1"
        return self.half_channels[direction]

    def __repr__(self):
        return "GossmapChannel[{}]".format(str(self.scid))

    def __str__(self):
        return str(self.scid)

    def __eq__(self, other):
        if not isinstance(other, GossmapChannel):
            return False
        return self.scid.__eq__(other.scid)

    def __hash__(self):
        return self.scid.__hash__()

    def has_feature(self, bit):
        return 3 << bit & self.features != 0

    def has_feature_compulsory(self, bit):
        return 1 << bit & self.features != 0

    def has_feature_optional(self, bit):
        return 2 << bit & self.features != 0

    def has_features(self, *bits):
        for bit in bits:
            if not self.has_feature(bit):
                return False
        return True

    def is_tor_only(c):
        """ Checks if a channel has TOR only nodes on both ends """
        return c.node1.is_tor_only() and c.node2.is_tor_only()


class GossmapNode(object):
    """A node: fields of node_announcement are in .fields,
       which can be None if there has been no node announcement.
       .channels is a list of the GossmapChannels attached to this node."""
    def __init__(self, node_id: Union[GossmapNodeId, bytes, str]):
        if isinstance(node_id, bytes) or isinstance(node_id, str):
            node_id = GossmapNodeId(node_id)
        self.fields: Optional[Dict[str, Any]] = None
        self.hdr: GossipStoreMsgHeader = None
        self.channels: List[GossmapChannel] = []
        self.node_id = node_id
        self.announced = False

        self._hash = self.node_id.__hash__()

    def __repr__(self):
        if hasattr(self, 'alias'):
            return f"GossmapNode[{self.node_id.nodeid.hex()}, \"{self.alias}\"]"
        return f"GossmapNode[{self.node_id.nodeid.hex()}]"

    def __eq__(self, other):
        if not isinstance(other, GossmapNode):
            return False
        return self.node_id.__eq__(other.node_id)

    def __lt__(self, other):
        if not isinstance(other, GossmapNode):
            raise TypeError(f"Cannot compare GossmapNode with {type(other)}")
        return self.node_id.__lt__(other.node_id)

    def __hash__(self):
        return self._hash

    def __str__(self):
        return str(self.node_id)

    def has_feature(self, bit):
        if not self.announced:
            return None
        return 3 << bit & self.features != 0

    def has_feature_compulsory(self, bit):
        if not self.announced:
            return None
        return 1 << bit & self.features != 0

    def has_feature_optional(self, bit):
        if not self.announced:
            return None
        return 2 << bit & self.features != 0

    def has_features(self, *bits):
        if not self.announced:
            return None
        for bit in bits:
            if not self.has_feature(bit):
                return False
        return True

    def _parse_addresses(self, data: bytes):
        """ parse address descriptors defined in bolts 07-routing-gossip.md """
        result = []
        try:
            stream = io.BytesIO(data)
            while stream.tell() < len(data):
                _type = int.from_bytes(stream.read(1), byteorder='big')
                if _type == 1:      # IPv4  length   6
                    ip = socket.inet_ntoa(stream.read(4))
                    port = int.from_bytes(stream.read(2), byteorder='big')
                    result.append(f"{ip}:{port}")
                elif _type == 2:    # IPv6  length  18
                    ip = socket.inet_ntop(socket.AF_INET6, stream.read(16))
                    port = int.from_bytes(stream.read(2), byteorder='big')
                    result.append(f"[{ip}]:{port}")
                elif _type == 3:    # TORv2 length  12 (deprecated)
                    stream.read(12)
                elif _type == 4:    # TORv3 length  37
                    addr = base64.b32encode(stream.read(35)).decode('ascii').lower()
                    port = int.from_bytes(stream.read(2), byteorder='big')
                    result.append(f"{addr}.onion:{port}")
                elif _type == 5:    # DNS   up to  258
                    hostname_len = int.from_bytes(stream.read(1), byteorder='big')
                    hostname = stream.read(hostname_len).decode('ascii')
                    port = int.from_bytes(stream.read(2), byteorder='big')
                    result.append(f"{hostname}:{port}")
                else:  # Stop parsing at the first unknown type
                    break
        # we simply pass exceptions and return what we were able to read so far
        except Exception:
            pass
        self.addresses = result

    def get_address_type(self, idx: int):
        """ I know this can be more sophisticated, but works """
        if not self.announced or len(self.addresses) <= idx:
            return None
        addrstr = self.addresses[idx]
        if ".onion:" in addrstr:
            return 'tor'
        if addrstr[0].isdigit():
            return 'ipv4'
        if addrstr.startswith("["):
            return 'ipv6'
        return 'dns'

    def has_clearnet(self):
        """ Checks if a node has one or more clearnet addresses """
        if not self.announced or len(self.addresses) == 0:
            return False
        for i in range(len(self.addresses)):
            if self.get_address_type(i) != 'tor':
                return True
        return False

    def has_tor(self):
        """ Checks if a node has one or more TOR addresses """
        if not self.announced or len(self.addresses) == 0:
            return False
        for i in range(len(self.addresses)):
            if self.get_address_type(i) == 'tor':
                return True
        return False

    def is_tor_only(self):
        """ Checks if a node has only TOR and no addresses announced """
        if not self.announced or len(self.addresses) == 0:
            return False
        for i in range(len(self.addresses)):
            if self.get_address_type(i) != 'tor':
                return False
        return True

    def is_tor_strict(self):
        """ Checks if a node is TOR only
            and is not publicly connected to any non-TOR nodes """
        if not self.is_tor_only():
            return False
        for c in self.channels:
            other = c.node1 if self != c.node1 else c.node2
            if other.has_tor():
                continue
            return False
        return True


class Gossmap(object):
    """Class to represent the gossip map of the network"""
    def __init__(self, store_filename: str = "gossip_store"):
        self.store_filename = store_filename
        self.store_file = open(store_filename, "rb")
        self.bytes_read = 1
        self.nodes: Dict[GossmapNodeId, GossmapNode] = {}
        self.channels: Dict[ShortChannelId, GossmapChannel] = {}
        self._last_scid: Optional[str] = None
        version = self.store_file.read(1)[0]
        if (version & GOSSIP_STORE_MAJOR_VERSION_MASK) != GOSSIP_STORE_MAJOR_VERSION:
            raise ValueError("Invalid gossip store version {}".format(version))
        self.processing_time = 0
        self.orphan_channel_updates = set()
        self.refresh()

    def _new_channel(self,
                     fields: Dict[str, Any],
                     scid: ShortChannelId,
                     node1: GossmapNode,
                     node2: GossmapNode,
                     is_private: bool,
                     hdr: GossipStoreMsgHeader):
        c = GossmapChannel(fields, scid, node1, node2, is_private, hdr)
        self._last_scid = scid
        self.channels[scid] = c
        node1.channels.append(c)
        node2.channels.append(c)

    def _del_channel(self, scid: ShortChannelId):
        c = self.channels[scid]
        del self.channels[scid]
        c.node1.channels.remove(c)
        c.node2.channels.remove(c)
        # Beware self-channels n1-n1!
        if len(c.node1.channels) == 0 and c.node1 != c.node2:
            del self.nodes[c.node1.node_id]
        if len(c.node2.channels) == 0:
            del self.nodes[c.node2.node_id]

    def _add_channel(self, rec: bytes, is_private: bool, hdr: GossipStoreMsgHeader):
        fields = channel_announcement.read(io.BytesIO(rec[2:]), {})
        # Add nodes one the fly
        node1_id = GossmapNodeId(fields['node_id_1'])
        node2_id = GossmapNodeId(fields['node_id_2'])
        if node1_id not in self.nodes:
            self.nodes[node1_id] = GossmapNode(node1_id)
        if node2_id not in self.nodes:
            self.nodes[node2_id] = GossmapNode(node2_id)
        self._new_channel(fields,
                          ShortChannelId.from_int(fields['short_channel_id']),
                          self.get_node(node1_id), self.get_node(node2_id),
                          is_private, hdr)

    def _set_channel_amount(self, rec: bytes):
        """ Sets channel capacity of last added channel """
        sats, = struct.unpack(">Q", rec[2:])
        self.channels[self._last_scid].satoshis = sats

    def get_channel(self, short_channel_id: Union[ShortChannelId, str]):
        """ Resolves a channel by its short channel id """
        if isinstance(short_channel_id, str):
            short_channel_id = ShortChannelId.from_str(short_channel_id)
        return self.channels.get(short_channel_id)

    def get_halfchannel(self,
                        short_channel_id: Union[ShortChannelId, str],
                        direction: int):
        """ Returns a GossmapHalfchannel identified by a scid and direction. """
        assert short_channel_id is not None
        if isinstance(short_channel_id, str):
            short_channel_id = ShortChannelId.from_str(short_channel_id)
        assert direction in [0, 1], "direction can only be 0 or 1"
        channel = self.get_channel(short_channel_id)
        return channel.half_channels[direction]

    def get_neighbors_hc(self,
                         source: Union[GossmapNodeId, str, None] = None,
                         destination: Union[GossmapNodeId, str, None] = None,
                         depth: int = 0,
                         excludes: Union[Set[Any], List[Any]] = set()):
        """ Returns a set[GossmapHalfchannel]` from `source` or towards
            `destination` node ID. Using the optional `depth` greater than `0`
            will result in a second, third, ...  order list of connected
            channels towards or from that node.
            Note: only one of `source` or `destination` can be given. """
        assert (source is None) ^ (destination is None), "Only one of source or destination must be given"
        assert depth >= 0, "Depth cannot be smaller than 0"
        node = self.get_node(source if source else destination)
        assert node is not None, "source or destination unknown"
        if isinstance(excludes, List):
            excludes = set(excludes)

        # first get set of reachable nodes ...
        reachable = self.get_neighbors(source, destination, depth, excludes)
        # and iterate and check any each source/dest channel from here
        result = set()
        for node in reachable:
            for channel in node.channels:
                if channel in excludes:
                    continue
                other = channel.node1 if node != channel.node1 else channel.node2
                if other in reachable or other in excludes:
                    continue
                direction = 0
                if source is not None and node > other:
                    direction = 1
                if destination is not None and node < other:
                    direction = 1
                hc = channel.half_channels[direction]
                # skip excluded or non existent halfchannels
                if hc is None or hc in excludes:
                    continue
                result.add(hc)
        return result

    def get_node(self, node_id: Union[GossmapNodeId, str]):
        """ Resolves a node by its public key node_id """
        if isinstance(node_id, str):
            node_id = GossmapNodeId.from_str(node_id)
        return self.nodes.get(node_id)

    def get_neighbors(self,
                      source: Union[GossmapNodeId, str, None] = None,
                      destination: Union[GossmapNodeId, str, None] = None,
                      depth: int = 0,
                      excludes: Union[Set[Any], List[Any]] = set()):
        """ Returns a set of nodes within a given depth from a source node """
        assert (source is None) ^ (destination is None), "Only one of source or destination must be given"
        assert depth >= 0, "Depth cannot be smaller than 0"
        node = self.get_node(source if source else destination)
        assert node is not None, "source or destination unknown"
        if isinstance(excludes, List):
            excludes = set(excludes)

        result = set()
        result.add(node)
        inner = set()
        inner.add(node)
        while depth > 0:
            shell = set()
            for node in inner:
                for channel in node.channels:
                    if channel in excludes:  # skip excluded channels
                        continue
                    other = channel.node1 if channel.node1 != node else channel.node2
                    direction = 0
                    if source is not None and node > other:
                        direction = 1
                    if destination is not None and node < other:
                        direction = 1
                    if channel.half_channels[direction] is None:
                        continue  # one way channel in the wrong direction
                    halfchannel = channel.half_channels[direction]
                    if halfchannel in excludes:  # skip excluded halfchannels
                        continue
                    # skip excluded or already seen nodes
                    if other in excludes or other in inner or other in result:
                        continue
                    shell.add(other)
            if len(shell) == 0:
                break
            depth -= 1
            result.update(shell)
            inner = shell
        return result

    def _update_channel(self, rec: bytes, hdr: GossipStoreMsgHeader):
        fields = channel_update.read(io.BytesIO(rec[2:]), {})
        direction = fields['channel_flags'] & 1
        scid = ShortChannelId.from_int(fields['short_channel_id'])
        if scid in self.channels:
            c = self.channels[scid]
            c._update_channel(direction, fields, hdr)
        else:
            self.orphan_channel_updates.add(scid)

    def _add_node_announcement(self, rec: bytes, hdr: GossipStoreMsgHeader):
        fields = node_announcement.read(io.BytesIO(rec[2:]), {})
        node_id = GossmapNodeId(fields['node_id'])
        if node_id not in self.nodes:
            self.nodes[node_id] = GossmapNode(node_id)
        node = self.nodes[node_id]
        node.fields = fields
        node.hdr = hdr

        # read metadata
        node.features = _parse_features(fields['features'])
        node.timestamp = fields['timestamp']
        node.alias = bytes(fields['alias']).decode('utf-8')
        node.rgb = fields['rgb_color']
        node._parse_addresses(bytes(fields['addresses']))
        node.announced = True

    def reopen_store(self):
        assert False, "FIXME: Implement!"

    def _remove_channel_by_deletemsg(self, rec: bytes):
        scidint, = struct.unpack(">Q", rec[2:])
        scid = ShortChannelId.from_int(scidint)
        # It might have already been deleted when we skipped it.
        if scid in self.channels:
            self._del_channel(scid)

    def _read_record(self) -> Optional[bytes]:
        """If a whole record is not in the file, returns None, None."""
        prev_off = self.bytes_read
        hdr = self.store_file.read(12)
        if len(hdr) != 12:
            self.store_file.seek(prev_off)
            return None, None
        hdr = GossipStoreMsgHeader(hdr, prev_off)
        rec = self.store_file.read(hdr.length)
        if len(rec) != hdr.length:
            self.store_file.seek(prev_off)
            return None, None
        if (hdr.flags & GOSSIP_STORE_LEN_COMPLETE_BIT) == 0:
            self.store_file.seek(prev_off)
            return None, None

        # Ok, we're digesting this one, so increment bytes_read.
        self.bytes_read += 12 + hdr.length
        return rec, hdr

    def refresh(self):
        """Catch up with any changes to the gossip store"""
        start_time = time.time()
        while True:
            rec, hdr = self._read_record()
            if rec is None:  # EOF
                break
            if hdr.deleted:  # Skip deleted records
                continue

            rectype, = struct.unpack(">H", rec[:2])
            if rectype == channel_announcement.number:
                self._add_channel(rec, False, hdr)
            elif rectype == WIRE_GOSSIP_STORE_PRIVATE_CHANNEL:
                hdr.off += 2 + 8 + 2
                self._add_channel(rec[2 + 8 + 2:], True, hdr)
            elif rectype == WIRE_GOSSIP_STORE_CHANNEL_AMOUNT:
                self._set_channel_amount(rec)
            elif rectype == channel_update.number:
                self._update_channel(rec, hdr)
            elif rectype == WIRE_GOSSIP_STORE_PRIVATE_UPDATE:
                hdr.off += 2 + 2
                self._update_channel(rec[2 + 2:], hdr)
            elif rectype == WIRE_GOSSIP_STORE_DELETE_CHAN:
                self._remove_channel_by_deletemsg(rec)
            elif rectype == node_announcement.number:
                self._add_node_announcement(rec, hdr)
            elif rectype == WIRE_GOSSIP_STORE_ENDED:
                self.reopen_store()
            else:
                continue
        self.processing_time += time.time() - start_time



from pyln.client import Millisatoshi
import pytest  # type: ignore


def test_to_approx_str():
    amount = Millisatoshi('10000000sat')
    assert amount.to_approx_str() == "0.1btc"
    amount = Millisatoshi('1000000sat')
    assert amount.to_approx_str() == "0.01btc"
    amount = Millisatoshi('100000sat')
    assert amount.to_approx_str() == "0.001btc"
    amount = Millisatoshi('10000sat')
    assert amount.to_approx_str() == "10000sat"
    amount = Millisatoshi('1000sat')
    assert amount.to_approx_str() == "1000sat"
    amount = Millisatoshi('100msat')
    assert amount.to_approx_str() == "0.1sat"

    # also test significant rounding
    amount = Millisatoshi('10001234sat')
    assert amount.to_approx_str() == "0.1btc"
    amount = Millisatoshi('1234sat')
    assert amount.to_approx_str(3) == "1234sat"  # note: no rounding
    amount = Millisatoshi('1234sat')
    assert amount.to_approx_str(2) == "1234sat"  # note: no rounding
    amount = Millisatoshi('1230sat')
    assert amount.to_approx_str(2) == "1230sat"  # note: no rounding
    amount = Millisatoshi('12345678sat')
    assert amount.to_approx_str() == "0.123btc"
    amount = Millisatoshi('12345678sat')
    assert amount.to_approx_str(1) == "0.1btc"
    amount = Millisatoshi('15345678sat')
    assert amount.to_approx_str(1) == "0.2btc"
    amount = Millisatoshi('1200000000sat')
    assert amount.to_approx_str() == "12btc"
    amount = Millisatoshi('1200000000sat')
    assert amount.to_approx_str(1) == "12btc"  # note: no rounding


def test_floats():
    # test parsing amounts from floating number strings
    amount = Millisatoshi("0.01btc")
    assert amount.to_satoshi() == 10**6
    amount = Millisatoshi("1.01btc")
    assert amount.to_satoshi() == 10**8 + 10**6
    amount = Millisatoshi("0.1sat")
    assert int(amount) == 100
    amount = Millisatoshi("0.01sat")
    assert int(amount) == 10
    amount = Millisatoshi("1.1sat")
    assert int(amount) == 1100

    # test floating point arithmetic
    amount = Millisatoshi("1000msat") * 0.1
    assert int(amount) == 100
    amount = Millisatoshi("100msat") * 0.1
    assert int(amount) == 10
    amount = Millisatoshi("10msat") * 0.1
    assert int(amount) == 1


def test_zero():
    # zero amounts are of course valid
    amount = Millisatoshi("0btc")
    assert int(amount) == 0
    amount = Millisatoshi("0sat")
    assert int(amount) == 0
    amount = Millisatoshi("0msat")
    assert int(amount) == 0

    # zero floating amount as well
    amount = Millisatoshi("0.0btc")
    assert int(amount) == 0
    amount = Millisatoshi("0.0sat")
    assert int(amount) == 0
    amount = Millisatoshi("0.0msat")
    assert int(amount) == 0

    # also anything multiplied by zero
    amount = Millisatoshi("1btc") * 0
    assert int(amount) == 0
    amount = Millisatoshi("1sat") * 0
    assert int(amount) == 0
    amount = Millisatoshi("1msat") * 0
    assert int(amount) == 0

    # and multiplied by a floating zero
    amount = Millisatoshi("1btc") * 0.0
    assert int(amount) == 0
    amount = Millisatoshi("1sat") * 0.0
    assert int(amount) == 0
    amount = Millisatoshi("1msat") * 0.0
    assert int(amount) == 0


def test_round_zero():
    # everything below 1msat should round down to zero
    amount = Millisatoshi("1msat") * 0.9
    assert int(amount) == 0
    amount = Millisatoshi("10msat") * 0.09
    assert int(amount) == 0
    amount = Millisatoshi("100msat") * 0.009
    assert int(amount) == 0
    amount = Millisatoshi("1000msat") * 0.0009
    assert int(amount) == 0

    amount = Millisatoshi("1sat") * 0.0009
    assert int(amount) == 0
    amount = Millisatoshi("0.1sat") * 0.009
    assert int(amount) == 0
    amount = Millisatoshi("0.01sat") * 0.09
    assert int(amount) == 0
    amount = Millisatoshi("0.001sat") * 0.9
    assert int(amount) == 0

    amount = Millisatoshi("10sat") * 0.00009
    assert int(amount) == 0
    amount = Millisatoshi("100sat") * 0.000009
    assert int(amount) == 0
    amount = Millisatoshi("1000sat") * 0.0000009
    assert int(amount) == 0
    amount = Millisatoshi("10000sat") * 0.00000009
    assert int(amount) == 0
    amount = Millisatoshi("10000sat") * 0.00000009
    assert int(amount) == 0

    amount = Millisatoshi("1btc") * 0.000000000009
    assert int(amount) == 0
    amount = Millisatoshi("0.1btc") * 0.00000000009
    assert int(amount) == 0
    amount = Millisatoshi("0.01btc") * 0.0000000009
    assert int(amount) == 0
    amount = Millisatoshi("0.001btc") * 0.000000009
    assert int(amount) == 0
    amount = Millisatoshi("0.0001btc") * 0.00000009
    assert int(amount) == 0
    amount = Millisatoshi("0.00001btc") * 0.0000009
    assert int(amount) == 0
    amount = Millisatoshi("0.000001btc") * 0.000009
    assert int(amount) == 0
    amount = Millisatoshi("0.0000001btc") * 0.00009
    assert int(amount) == 0
    amount = Millisatoshi("0.00000001btc") * 0.0009
    assert int(amount) == 0
    amount = Millisatoshi("0.000000001btc") * 0.009
    assert int(amount) == 0
    amount = Millisatoshi("0.0000000001btc") * 0.09
    assert int(amount) == 0
    amount = Millisatoshi("0.00000000001btc") * 0.9
    assert int(amount) == 0


def test_round_down():
    # sub msat significatns should be floored
    amount = Millisatoshi("2msat") * 0.9
    assert int(amount) == 1
    amount = Millisatoshi("20msat") * 0.09
    assert int(amount) == 1
    amount = Millisatoshi("200msat") * 0.009
    assert int(amount) == 1
    amount = Millisatoshi("2000msat") * 0.0009
    assert int(amount) == 1

    amount = Millisatoshi("2sat") * 0.0009
    assert int(amount) == 1
    amount = Millisatoshi("0.2sat") * 0.009
    assert int(amount) == 1
    amount = Millisatoshi("0.02sat") * 0.09
    assert int(amount) == 1
    amount = Millisatoshi("0.002sat") * 0.9
    assert int(amount) == 1

    amount = Millisatoshi("20sat") * 0.00009
    assert int(amount) == 1
    amount = Millisatoshi("200sat") * 0.000009
    assert int(amount) == 1
    amount = Millisatoshi("2000sat") * 0.0000009
    assert int(amount) == 1
    amount = Millisatoshi("20000sat") * 0.00000009
    assert int(amount) == 1
    amount = Millisatoshi("20000sat") * 0.00000009
    assert int(amount) == 1

    amount = Millisatoshi("2btc") * 0.000000000009
    assert int(amount) == 1
    amount = Millisatoshi("0.2btc") * 0.00000000009
    assert int(amount) == 1
    amount = Millisatoshi("0.02btc") * 0.0000000009
    assert int(amount) == 1
    amount = Millisatoshi("0.002btc") * 0.000000009
    assert int(amount) == 1
    amount = Millisatoshi("0.0002btc") * 0.00000009
    assert int(amount) == 1
    amount = Millisatoshi("0.00002btc") * 0.0000009
    assert int(amount) == 1
    amount = Millisatoshi("0.000002btc") * 0.000009
    assert int(amount) == 1
    amount = Millisatoshi("0.0000002btc") * 0.00009
    assert int(amount) == 1
    amount = Millisatoshi("0.00000002btc") * 0.0009
    assert int(amount) == 1
    amount = Millisatoshi("0.000000002btc") * 0.009
    assert int(amount) == 1
    amount = Millisatoshi("0.0000000002btc") * 0.09
    assert int(amount) == 1
    amount = Millisatoshi("0.00000000002btc") * 0.9
    assert int(amount) == 1


def test_nosubmsat():
    # sub millisatoshi are not a concept yet
    with pytest.raises(ValueError, match='Millisatoshi must be a whole number'):
        Millisatoshi("0.1msat")
    with pytest.raises(ValueError, match='Millisatoshi must be a whole number'):
        Millisatoshi(".1msat")
    with pytest.raises(ValueError, match='Millisatoshi must be a whole number'):
        Millisatoshi("0.0001sat")
    with pytest.raises(ValueError, match='Millisatoshi must be a whole number'):
        Millisatoshi(".0001sat")
    with pytest.raises(ValueError, match='Millisatoshi must be a whole number'):
        Millisatoshi("0.000000000001btc")
    with pytest.raises(ValueError, match='Millisatoshi must be a whole number'):
        Millisatoshi(".000000000001btc")


def test_nonegative():
    with pytest.raises(ValueError, match='Millisatoshi must be >= 0'):
        Millisatoshi("-1btc")
    with pytest.raises(ValueError, match='Millisatoshi must be >= 0'):
        Millisatoshi("-1.0btc")
    with pytest.raises(ValueError, match='Millisatoshi must be >= 0'):
        Millisatoshi("-0.1btc")
    with pytest.raises(ValueError, match='Millisatoshi must be >= 0'):
        Millisatoshi("-.1btc")
    with pytest.raises(ValueError, match='Millisatoshi must be >= 0'):
        Millisatoshi("-1sat")
    with pytest.raises(ValueError, match='Millisatoshi must be >= 0'):
        Millisatoshi("-1.0sat")
    with pytest.raises(ValueError, match='Millisatoshi must be >= 0'):
        Millisatoshi("-0.1sat")
    with pytest.raises(ValueError, match='Millisatoshi must be >= 0'):
        Millisatoshi("-.1sat")
    with pytest.raises(ValueError, match='Millisatoshi must be >= 0'):
        Millisatoshi("-1msat")
    with pytest.raises(ValueError, match='Millisatoshi must be >= 0'):
        Millisatoshi("-1.0msat")

    with pytest.raises(ValueError, match='Millisatoshi must be >= 0'):
        Millisatoshi("1msat") * -1
    with pytest.raises(ValueError, match='Millisatoshi must be >= 0'):
        Millisatoshi("1msat") * -42
    with pytest.raises(ValueError, match='Millisatoshi must be >= 0'):
        Millisatoshi("1sat") * -1
    with pytest.raises(ValueError, match='Millisatoshi must be >= 0'):
        Millisatoshi("1btc") * -1

    with pytest.raises(ValueError, match='Millisatoshi must be >= 0'):
        Millisatoshi("1msat") / -1
    with pytest.raises(ValueError, match='Millisatoshi must be >= 0'):
        Millisatoshi("1msat") / -0.5
    with pytest.raises(ValueError, match='Millisatoshi must be >= 0'):
        Millisatoshi("1sat") / -1
    with pytest.raises(ValueError, match='Millisatoshi must be >= 0'):
        Millisatoshi("1btc") / -1

    with pytest.raises(ValueError, match='Millisatoshi must be >= 0'):
        Millisatoshi("1msat") // -1
    with pytest.raises(ValueError, match='Millisatoshi must be >= 0'):
        Millisatoshi("1sat") // -1
    with pytest.raises(ValueError, match='Millisatoshi must be >= 0'):
        Millisatoshi("1btc") // -1


def test_mul():
    # msat * num := msat
    amount = Millisatoshi(21) * 2
    assert isinstance(amount, Millisatoshi)
    assert amount == Millisatoshi(42)
    amount = Millisatoshi(21) * 2.5
    assert amount == Millisatoshi(52)

    # msat * msat := msat^2  (which is not supported)
    with pytest.raises(TypeError, match="not supported"):
        Millisatoshi(21) * Millisatoshi(2)


def test_div():
    # msat / num := msat
    amount = Millisatoshi(42) / 2
    assert isinstance(amount, Millisatoshi)
    assert amount == Millisatoshi(21)
    amount = Millisatoshi(42) / 2.6
    assert amount == Millisatoshi(16)

    # msat / msat := num   (float ratio)
    amount = Millisatoshi(42) / Millisatoshi(2)
    assert isinstance(amount, float)
    assert amount == 21.0
    amount = Millisatoshi(8) / Millisatoshi(5)
    assert amount == 1.6

    # msat // num := msat
    amount = Millisatoshi(42) // 2
    assert isinstance(amount, Millisatoshi)
    assert amount == Millisatoshi(21)

    # msat // msat := num
    amount = Millisatoshi(42) // Millisatoshi(3)
    assert isinstance(amount, int)
    assert amount == 14
    amount = Millisatoshi(42) // Millisatoshi(3)
    assert amount == 14
    amount = Millisatoshi(42) // Millisatoshi(4)
    assert amount == 10


def test_init():
    # Note: Ongoing Discussion, hence the `with pytest.raises`.
    # https://github.com/ElementsProject/lightning/pull/4273#discussion_r540369093
    #
    # Initialization with a float should be possible:
    # Millisatoshi(5) / 2 currently works, and removes the half msat.
    # So Millisatoshi(5 / 2) should be the same.
    amount = Millisatoshi(5) / 2
    assert amount == Millisatoshi(2)
    with pytest.raises(TypeError, match="Millisatoshi by float is currently not supported"):
        assert amount == Millisatoshi(5 / 2)

    ratio = Millisatoshi(8) / Millisatoshi(5)
    assert isinstance(ratio, float)
    with pytest.raises(TypeError, match="Millisatoshi by float is currently not supported"):
        assert Millisatoshi(ratio) == Millisatoshi(8 / 5)

    # Check that init by a round float is allowed.
    # Required by some existing tests: tests/test_wallet.py::test_txprepare
    amount = Millisatoshi(42.0)
    assert amount == 42



from pyln.client import Plugin
from pyln.client.plugin import Request, Millisatoshi, RpcException
import itertools
import pytest  # type: ignore


def test_simple_methods():
    """Test the dispatch of methods, with a variety of bindings.
    """
    call_list = []
    p = Plugin(autopatch=False)

    @p.method("test1")
    def test1(name):
        """Has a single positional argument."""
        assert name == 'World'
        call_list.append(test1)
    request = p._parse_request({
        'id': 1,
        'jsonrpc': '2.0',
        'method': 'test1',
        'params': {'name': 'World'}
    })
    p._dispatch_request(request)
    assert call_list == [test1]

    @p.method("test2")
    def test2(name, plugin):
        """Also asks for the plugin instance. """
        assert plugin == p
        call_list.append(test2)
    request = p._parse_request({
        'id': 1,
        'jsonrpc': '2.0',
        'method': 'test2',
        'params': {'name': 'World'}
    })
    p._dispatch_request(request)
    assert call_list == [test1, test2]

    @p.method("test3")
    def test3(name, request):
        """Also asks for the request instance. """
        assert request is not None
        call_list.append(test3)
    request = p._parse_request({
        'id': 1,
        'jsonrpc': '2.0',
        'method': 'test3',
        'params': {'name': 'World'}
    })
    p._dispatch_request(request)
    assert call_list == [test1, test2, test3]

    @p.method("test4")
    def test4(name):
        """Try the positional arguments."""
        assert name == 'World'
        call_list.append(test4)
    request = p._parse_request({
        'id': 1,
        'jsonrpc': '2.0',
        'method': 'test4',
        'params': ['World']
    })
    p._dispatch_request(request)
    assert call_list == [test1, test2, test3, test4]

    @p.method("test5")
    def test5(name, request, plugin):
        """Try the positional arguments, mixing in the request and plugin."""
        assert name == 'World'
        assert request is not None
        assert p == plugin
        call_list.append(test5)
    request = p._parse_request({
        'id': 1,
        'jsonrpc': '2.0',
        'method': 'test5',
        'params': ['World']
    })
    p._dispatch_request(request)
    assert call_list == [test1, test2, test3, test4, test5]

    answers = []

    @p.method("test6")
    def test6(name, answer=42):
        """This method has a default value for one of its params"""
        assert name == 'World'
        answers.append(answer)
        call_list.append(test6)

    # Both calls should work (with and without the default param
    request = p._parse_request({
        'id': 1,
        'jsonrpc': '2.0',
        'method': 'test6',
        'params': ['World']
    })
    p._dispatch_request(request)
    assert call_list == [test1, test2, test3, test4, test5, test6]
    assert answers == [42]

    request = p._parse_request({
        'id': 1,
        'jsonrpc': '2.0',
        'method': 'test6',
        'params': ['World', 31337]
    })
    p._dispatch_request(request)
    assert call_list == [test1, test2, test3, test4, test5, test6, test6]
    assert answers == [42, 31337]


def test_methods_errors():
    """A bunch of tests that should fail calling the methods."""
    call_list = []
    p = Plugin(autopatch=False)

    # Fails because we haven't added the method yet
    request = p._parse_request({
        'id': 1,
        'jsonrpc': '2.0',
        'method': 'test1',
        'params': {}
    })
    with pytest.raises(ValueError):
        p._dispatch_request(request)
    assert call_list == []

    @p.method("test1")
    def test1(name):
        call_list.append(test1)

    # Attempting to add it twice should fail
    with pytest.raises(ValueError):
        p.add_method("test1", test1)

    # Fails because it is missing the 'name' argument
    request = p._parse_request({'id': 1, 'jsonrpc': '2.0', 'method': 'test1', 'params': {}})
    with pytest.raises(TypeError):
        p._exec_func(test1, request)
    assert call_list == []

    # The same with positional arguments
    request = p._parse_request({'id': 1, 'jsonrpc': '2.0', 'method': 'test1', 'params': []})
    with pytest.raises(TypeError):
        p._exec_func(test1, request)
    assert call_list == []

    # Fails because we have a non-matching argument
    request = p._parse_request({
        'id': 1,
        'jsonrpc': '2.0',
        'method': 'test1',
        'params': {'name': 'World', 'extra': 1}
    })
    with pytest.raises(TypeError):
        p._exec_func(test1, request)
    assert call_list == []

    request = p._parse_request({
        'id': 1,
        'jsonrpc': '2.0',
        'method': 'test1',
        'params': ['World', 1]
    })

    with pytest.raises(TypeError):
        p._exec_func(test1, request)
    assert call_list == []


def test_method_exceptions():
    """A bunch of tests that should fail calling the methods."""
    p = Plugin(autopatch=False)

    def fake_write_result(resultdict):
        global result_dict
        result_dict = resultdict

    @p.method("test_raise")
    def test_raise():
        raise RpcException("testing RpcException", code=-1000)

    req = Request(p, 1, "test_raise", {})
    req._write_result = fake_write_result
    p._dispatch_request(req)
    assert result_dict['jsonrpc'] == '2.0'
    assert result_dict['id'] == 1
    assert result_dict['error']['code'] == -1000
    assert result_dict['error']['message'] == "testing RpcException"

    @p.method("test_raise2")
    def test_raise2():
        raise Exception("normal exception")

    req = Request(p, 1, "test_raise2", {})
    req._write_result = fake_write_result
    p._dispatch_request(req)
    assert result_dict['jsonrpc'] == '2.0'
    assert result_dict['id'] == 1
    assert result_dict['error']['code'] == -32600
    assert result_dict['error']['message'] == "Error while processing test_raise2: normal exception"


def test_positional_inject():
    p = Plugin()
    rdict = Request(
        plugin=p,
        req_id=1,
        method='func',
        params={'a': 1, 'b': 2, 'kwa': 3, 'kwb': 4}
    )
    rarr = Request(
        plugin=p,
        req_id=1,
        method='func',
        params=[1, 2, 3, 4],
    )

    def pre_args(plugin, a, b, kwa=3, kwb=4):
        assert (plugin, a, b, kwa, kwb) == (p, 1, 2, 3, 4)

    def in_args(a, plugin, b, kwa=3, kwb=4):
        assert (plugin, a, b, kwa, kwb) == (p, 1, 2, 3, 4)

    def post_args(a, b, plugin, kwa=3, kwb=4):
        assert (plugin, a, b, kwa, kwb) == (p, 1, 2, 3, 4)

    def post_kwargs(a, b, kwa=3, kwb=4, plugin=None):
        assert (plugin, a, b, kwa, kwb) == (p, 1, 2, 3, 4)

    def in_multi_args(a, request, plugin, b, kwa=3, kwb=4):
        assert request in [rarr, rdict]
        assert (plugin, a, b, kwa, kwb) == (p, 1, 2, 3, 4)

    def in_multi_mix_args(a, plugin, b, request=None, kwa=3, kwb=4):
        assert request in [rarr, rdict]
        assert (plugin, a, b, kwa, kwb) == (p, 1, 2, 3, 4)

    def extra_def_arg(a, b, c, d, e=42):
        """ Also uses a different name for kwa and kwb
        """
        assert (a, b, c, d, e) == (1, 2, 3, 4, 42)

    def count(plugin, count, request):
        assert count == 42 and plugin == p

    funcs = [pre_args, in_args, post_args, post_kwargs, in_multi_args]

    for func, request in itertools.product(funcs, [rdict, rarr]):
        p._exec_func(func, request)

    p._exec_func(extra_def_arg, rarr)

    p._exec_func(count, Request(
        plugin=p,
        req_id=1,
        method='func',
        params=[42],
    ))

    # This should fail since it is missing one positional argument
    with pytest.raises(TypeError):
        p._exec_func(count, Request(
            plugin=p,
            req_id=1,
            method='func',
            params=[])
        )


def test_bind_pos():
    p = Plugin(autopatch=False)

    req = object()
    params = ['World']

    def test1(name):
        assert name == 'World'
    bound = p._bind_pos(test1, params, req)
    test1(*bound.args, **bound.kwargs)

    def test2(name, plugin):
        assert name == 'World'
        assert plugin == p
    bound = p._bind_pos(test2, params, req)
    test2(*bound.args, **bound.kwargs)

    def test3(plugin, name):
        assert name == 'World'
        assert plugin == p
    bound = p._bind_pos(test3, params, req)
    test3(*bound.args, **bound.kwargs)

    def test4(plugin, name, request):
        assert name == 'World'
        assert plugin == p
        assert request == req
    bound = p._bind_pos(test4, params, req)
    test4(*bound.args, **bound.kwargs)

    def test5(request, name, plugin):
        assert name == 'World'
        assert plugin == p
        assert request == req
    bound = p._bind_pos(test5, params, req)
    test5(*bound.args, **bound.kwargs)

    def test6(request, name, plugin, answer=42):
        assert name == 'World'
        assert plugin == p
        assert request == req
        assert answer == 42
    bound = p._bind_pos(test6, params, req)
    test6(*bound.args, **bound.kwargs)

    # Now mix in a catch-all parameter that needs to be assigned
    def test6(request, name, plugin, *args, **kwargs):
        assert name == 'World'
        assert plugin == p
        assert request == req
        assert args == (42,)
        assert kwargs == {}
    bound = p._bind_pos(test6, params + [42], req)
    test6(*bound.args, **bound.kwargs)


def test_bind_kwargs():
    p = Plugin(autopatch=False)

    req = object()
    params = {'name': 'World'}

    def test1(name):
        assert name == 'World'
    bound = p._bind_kwargs(test1, params, req)
    test1(*bound.args, **bound.kwargs)

    def test2(name, plugin):
        assert name == 'World'
        assert plugin == p
    bound = p._bind_kwargs(test2, params, req)
    test2(*bound.args, **bound.kwargs)

    def test3(plugin, name):
        assert name == 'World'
        assert plugin == p
    bound = p._bind_kwargs(test3, params, req)
    test3(*bound.args, **bound.kwargs)

    def test4(plugin, name, request):
        assert name == 'World'
        assert plugin == p
        assert request == req
    bound = p._bind_kwargs(test4, params, req)
    test4(*bound.args, **bound.kwargs)

    def test5(request, name, plugin):
        assert name == 'World'
        assert plugin == p
        assert request == req
    bound = p._bind_kwargs(test5, params, req)
    test5(*bound.args, **bound.kwargs)

    def test6(request, name, plugin, answer=42):
        assert name == 'World'
        assert plugin == p
        assert request == req
        assert answer == 42
    bound = p._bind_kwargs(test6, params, req)
    test6(*bound.args, **bound.kwargs)

    # Now mix in a catch-all parameter that needs to be assigned
    def test6(request, name, plugin, *args, **kwargs):
        assert name == 'World'
        assert plugin == p
        assert request == req
        assert args == ()
        assert kwargs == {'answer': 42}
    bound = p._bind_kwargs(test6, {'name': 'World', 'answer': 42}, req)
    test6(*bound.args, **bound.kwargs)


def test_argument_coercion():
    p = Plugin(autopatch=False)

    def test1(msat: Millisatoshi):
        assert isinstance(msat, Millisatoshi)

    ba = p._bind_kwargs(test1, {"msat": "100msat"}, None)
    test1(*ba.args)

    ba = p._bind_pos(test1, ["100msat"], None)
    test1(*ba.args, **ba.kwargs)


def test_duplicate_result():
    p = Plugin(autopatch=False)

    def test1(request):
        request.set_result(1)     # MARKER1
        request.set_result(1)

    req = Request(p, req_id=1, method="test1", params=[])
    ba = p._bind_kwargs(test1, {}, req)
    with pytest.raises(ValueError, match=r'current state is RequestState\.FINISHED(.*\n.*)*MARKER1'):
        test1(*ba.args)

    def test2(request):
        request.set_exception(1)  # MARKER2
        request.set_exception(1)

    req = Request(p, req_id=2, method="test2", params=[])
    ba = p._bind_kwargs(test2, {}, req)
    with pytest.raises(ValueError, match=r'current state is RequestState\.FAILED(.*\n*.*)*MARKER2'):
        test2(*ba.args)

    def test3(request):
        request.set_exception(1)  # MARKER3
        request.set_result(1)

    req = Request(p, req_id=3, method="test3", params=[])
    ba = p._bind_kwargs(test3, {}, req)
    with pytest.raises(ValueError, match=r'current state is RequestState\.FAILED(.*\n*.*)*MARKER3'):
        test3(*ba.args)

    def test4(request):
        request.set_result(1)     # MARKER4
        request.set_exception(1)

    req = Request(p, req_id=4, method="test4", params=[])
    ba = p._bind_kwargs(test4, {}, req)
    with pytest.raises(ValueError, match=r'current state is RequestState\.FINISHED(.*\n*.*)*MARKER4'):
        test4(*ba.args)


def test_usage():
    p = Plugin(autopatch=False)

    @p.method("some_method")
    def some_method(some_arg: str = None):
        """some description"""
        pass

    manifest = p._getmanifest()
    usage = p.get_usage()

    assert manifest['rpcmethods'][0]['name'] == 'some_method'
    assert "some_arg" in manifest['rpcmethods'][0]['usage']
    assert "some description" in manifest['rpcmethods'][0]['usage']
    assert "some_method" in usage
    assert "some_arg" in usage
    assert "some description" in usage



import pytest
from pyln.client import Millisatoshi


def test_sum_radd():
    result = sum([Millisatoshi(1), Millisatoshi(2), Millisatoshi(3)])
    assert int(result) == 6


def test_compare_int():
    # Test that we can compare msat to int numbers
    assert Millisatoshi(10) == 10
    assert Millisatoshi(10) > 9
    assert Millisatoshi(10) >= 9
    assert Millisatoshi(10) < 11
    assert Millisatoshi(10) <= 11

    # Same as above but check that the order doesn't matter
    assert 10 == Millisatoshi(10)
    assert 9 < Millisatoshi(10)
    assert 9 <= Millisatoshi(10)
    assert 11 > Millisatoshi(10)
    assert 11 >= Millisatoshi(10)

    # Test that we can't accidentally compare msat to float
    assert Millisatoshi(10) != 10.0
    with pytest.raises(AttributeError):
        assert Millisatoshi(10) > 9.0
    with pytest.raises(AttributeError):
        assert Millisatoshi(10) >= 9.0
    with pytest.raises(AttributeError):
        assert Millisatoshi(10) < 11.0
    with pytest.raises(AttributeError):
        assert Millisatoshi(10) <= 11.0

    # ... and again that order does not matter
    assert 10.0 != Millisatoshi(10)
    with pytest.raises(AttributeError):
        assert 9.0 < Millisatoshi(10)
    with pytest.raises(AttributeError):
        assert 9.0 <= Millisatoshi(10)
    with pytest.raises(AttributeError):
        assert 11.0 > Millisatoshi(10)
    with pytest.raises(AttributeError):
        assert 11.0 >= Millisatoshi(10)



from pyln.client.clnutils import cln_parse_rpcversion


def test_rpcversion():
    foo = cln_parse_rpcversion("0.11.2")
    assert foo[0] == 0
    assert foo[1] == 11
    assert foo[2] == 2

    foo = cln_parse_rpcversion("0.11.2rc2-modded")
    assert foo[0] == 0
    assert foo[1] == 11
    assert foo[2] == 2

    foo = cln_parse_rpcversion("22.11")
    assert foo[0] == 22
    assert foo[1] == 11
    assert foo[2] == 0

    foo = cln_parse_rpcversion("22.11rc1")
    assert foo[0] == 22
    assert foo[1] == 11
    assert foo[2] == 0

    foo = cln_parse_rpcversion("22.11rc1-modded")
    assert foo[0] == 22
    assert foo[1] == 11
    assert foo[2] == 0

    foo = cln_parse_rpcversion("22.11-modded")
    assert foo[0] == 22
    assert foo[1] == 11
    assert foo[2] == 0

    foo = cln_parse_rpcversion("22.11.0")
    assert foo[0] == 22
    assert foo[1] == 11
    assert foo[2] == 0

    foo = cln_parse_rpcversion("22.11.1")
    assert foo[0] == 22
    assert foo[1] == 11
    assert foo[2] == 1



from __future__ import annotations

from functools import total_ordering
from typing import Union


@total_ordering
class NodeVersion:
    """NodeVersion

    NodeVersion represents a Core Lightning Version. It also contains a scheme
    to compare versions against each-other.

    The main use-case to compare versions is to test for the availability of a feature.
    An example is the `--developer` flag which was introduced in cln v23.11

    We define the following rules
    - `v23.11` == `v23.11`
       Obviously a version is equal to it-self.
    - `v23.11rc3` == `v23.11rc1` == `v23.11`
      We don't add new features in release-candidates and therefore the feature-set is the same.
      This is also more ergonomic for a dev that wants to know if the `--developer` flag is available.
      See `strict_equal` if an exact match is required
    - `v23.11` < `v24.02`
      The oldest version is the smallest
    - `vd6fa78c`
      This is an untagged version, such as in CI.  This is assumed to be the latest, greater than
      any test.
    """
    def __init__(self, version: str):
        # e.g. v24.11-225-gda793e66b9
        if version.startswith('v'):
            version = version[1:]
        version = version.split('-')[0]
        parts = version.split('.')
        # rc is considered "close enough"
        if 'rc' in parts[-1]:
            parts[-1] = parts[-1].split('rc')[0]

        self.parts: int = []

        # Single part?  It's a git version, so treat it as the future.
        if len(parts) == 1:
            self.parts.append(100)
        else:
            for p in parts:
                self.parts.append(int(p))

    def __eq__(self, other: Union[NodeVersion, str]) -> bool:
        if isinstance(other, str):
            other = NodeVersion(other)
        if not isinstance(other, NodeVersion):
            return False

        if len(self.parts) != len(other.parts):
            return False
        for a, b in zip(self.parts, other.parts):
            if a != b:
                return False
        return True

    def __lt__(self, other: Union[NodeVersion, str]) -> bool:
        if isinstance(other, str):
            other = NodeVersion(other)
        if not isinstance(other, NodeVersion):
            return NotImplemented

        # We want a zero-padded zip.  Pad both to make one.
        totlen = max(len(self.parts), len(other.parts))
        for a, b in zip(self.parts + [0] * totlen, other.parts + [0] * totlen):
            if a < b:
                return True
            if a > b:
                return False
        return False


__all__ = [NodeVersion]



from pyln.client import Gossmap, GossmapNode, GossmapNodeId

import os.path
import lzma


def unxz_data_tmp(src, tmp_path, dst, wmode):
    fulldst = os.path.join(tmp_path, dst)
    with open(fulldst, wmode) as out:
        with lzma.open(os.path.join(os.path.dirname(__file__), "data", src), "rb")as f:
            out.write(f.read())
    return fulldst


def test_gossmap(tmp_path):
    sfile = unxz_data_tmp("gossip_store-part1.xz", tmp_path, "gossip_store", "xb")
    g = Gossmap(sfile)

    chans = len(g.channels)
    nodes = len(g.nodes)

    g.refresh()
    assert chans == len(g.channels)
    assert nodes == len(g.nodes)

    # Now append.
    unxz_data_tmp("gossip_store-part2.xz", tmp_path, "gossip_store", "ab")

    g.refresh()

    # This actually deletes a channel, which deletes a node.
    assert g.get_channel("686386x1093x1") is None
    assert g.get_node('029deaf9d2fba868fe0a124050f0a13e021519a12f41bea34f391fe7533fb3166d') is None
    # The other node is untouched
    assert g.get_node('02e0af3c70bf42343316513e54683b10c01d906c04a05dfcd9479b90d7beed9129')

    # It will notice the new ones.
    assert chans < len(g.channels)
    assert nodes < len(g.nodes)

    # Whole load at the same time gives the same results.
    g2 = Gossmap(sfile)
    assert set(g.channels.keys()) == set(g2.channels.keys())
    assert set(g.nodes.keys()) == set(g2.nodes.keys())

    # Check some details
    channel2 = g.get_channel("686200x1137x0")
    assert g.get_channel("686386x1093x1") is None
    assert channel2.satoshis == 3000000


def test_gossmap_halfchannel(tmp_path):
    """ this test a simple [l1->l2] gossip store that was created by the pyln-testing framework (tests/test_gossip.py::test_announce_address) """
    sfile = unxz_data_tmp("gossip_store.simple.xz", tmp_path, "gossip_store", "xb")
    g = Gossmap(sfile)

    l1id = "022d223620a359a47ff7f7ac447c85c46c923da53389221a0054c11c1e3ca31d59"
    l2id = "0266e4598d1d3c415f572a8488830b60f7e744ed9235eb0b1ba93283b315c03518"

    # check structure parsed correctly
    assert len(g.nodes) == 2
    n1 = g.get_node(l1id)
    n2 = g.get_node(l2id)
    assert n1
    assert n2

    chan = g.get_channel("103x1x0")
    assert chan
    assert chan.node1 == n1
    assert chan.node2 == n2

    half0 = chan.get_direction(0)
    half1 = g.get_halfchannel("103x1x0", 1)
    assert half0
    assert half1
    assert half0.direction == 0
    assert half1.direction == 1
    assert half0.channel == chan
    assert half1.channel == chan
    assert half0.source == n1
    assert half0.destination == n2
    assert half1.source == n2
    assert half1.destination == n1

    # check metadata
    assert half0.timestamp == 1759208366
    assert half1.timestamp == 1759208366
    assert half0.cltv_expiry_delta == 6
    assert half1.cltv_expiry_delta == 6
    assert half0.htlc_minimum_msat == 0
    assert half1.htlc_minimum_msat == 0
    assert half0.htlc_maximum_msat == 990000000
    assert half1.htlc_maximum_msat == 990000000
    assert half0.fee_base_msat == 1
    assert half1.fee_base_msat == 1
    assert half0.fee_proportional_millionths == 10


def test_objects():
    boltz = "026165850492521f4ac8abd9bd8088123446d126f648ca35e60f88177dc149ceb2"
    acinq = "03864ef025fde8fb587d989186ce6a4a186895ee44a926bfc370e2c366597a3f8f"

    boltz_id = GossmapNodeId(bytes.fromhex(boltz))
    acinq_id = GossmapNodeId(bytes.fromhex(acinq))
    assert boltz_id == GossmapNodeId(boltz)

    assert boltz_id < acinq_id
    assert acinq_id > boltz_id
    assert boltz_id != acinq_id
    assert acinq_id != boltz_id
    assert not boltz_id > acinq_id
    assert not acinq_id < boltz_id
    assert not boltz_id == acinq_id
    assert not acinq_id == boltz_id

    boltz_node = GossmapNode(boltz_id)
    acinq_node = GossmapNode(acinq_id)
    assert boltz_node == GossmapNode(boltz)
    assert boltz_node < acinq_node
    assert acinq_node > boltz_node
    assert boltz_node != acinq_node


def test_mesh(tmp_path):
    """This gossip store is a nice mesh created with pyln-testing:

       l1--l2--l3
       |   |   |
       l4--l5--l6
       |   |   |
       l7--l8--l9
    """
    sfile = unxz_data_tmp("gossip_store.mesh-3x3.xz", tmp_path, "gossip_store", "xb")
    g = Gossmap(sfile)
    assert len(g.nodes) == 9
    assert len(g.channels) == 12

    nodeids = ['0266e4598d1d3c415f572a8488830b60f7e744ed9235eb0b1ba93283b315c03518',
               '022d223620a359a47ff7f7ac447c85c46c923da53389221a0054c11c1e3ca31d59',
               '035d2b1192dfba134e10e540875d366ebc8bc353d5aa766b80c090b39c3a5d885d',
               '0382ce59ebf18be7d84677c2e35f23294b9992ceca95491fcf8a56c6cb2d9de199',
               '032cf15d1ad9c4a08d26eab1918f732d8ef8fdc6abb9640bf3db174372c491304e',
               '0265b6ab5ec860cd257865d61ef0bbf5b3339c36cbda8b26b74e7f1dca490b6518',
               '0269f9862c311261241e5aee7abe0ec93c88613cc8f3c5f33cb1eea90d2bc4ddb6',
               '03a7fd8070eea99341418fefe0b31086054d09cff64649eec3605db2340631c616',
               '030eeb52087b9dbb27b7aec79ca5249369f6ce7b20a5684ce38d9f4595a21c2fda']
    scid12 = '103x1x0'
    scid14 = '105x1x1'
    scid23 = '107x1x1'
    scid25 = '109x1x1'
    scid36 = '111x1x0'
    scid45 = '113x1x0'
    scid47 = '115x1x1'
    scid56 = '117x1x1'
    scid58 = '119x1x0'
    scid69 = '121x1x1'
    scid78 = '123x1x1'
    scid89 = '125x1x1'
    scids = [scid12, scid14, scid23, scid25, scid36, scid45, scid47, scid56,
             scid58, scid69, scid78, scid89]

    nodes = [g.get_node(nid) for nid in nodeids]

    # check all nodes are there
    for nodeid in nodeids:
        node = g.get_node(nodeid)
        assert node
        assert str(node.node_id) == nodeid
        for channel in node.channels:
            assert str(channel.scid) in scids

    # assert all channels are there
    for scid in scids:
        channel = g.get_channel(scid)
        assert channel
        assert str(channel.scid) == scid
        assert channel.half_channels[0]
        assert channel.half_channels[1]

    # check basic relations
    # get_neighbors l5 in the middle depth=0 returns just that node
    result = g.get_neighbors(source=nodeids[4])
    assert len(result) == 1
    assert str(next(iter(result)).node_id) == nodeids[4]
    result = g.get_neighbors(source=nodeids[4], depth=1)
    assert len(result) == 5
    # on depth=1 the cross l2, l4, l5, l6, l8 must be returned
    assert nodes[1] in result
    assert nodes[3] in result
    assert nodes[4] in result
    assert nodes[5] in result
    assert nodes[7] in result
    # on depth>=2 all nodes must be returned as we visited the whole graph
    for d in range(2, 4):
        result = g.get_neighbors(source=nodeids[4], depth=d)
        assert len(result) == 9
        for node in nodes:
            assert node in result
    # get_neighbors on l9 with depth=3 must return all but l1
    result = g.get_neighbors(nodeids[8], depth=3)
    assert len(result) == 8
    assert nodes[0] not in result
    # get_neighbors on l9 with depth=4 and excludes l5 must return all but l5
    result = g.get_neighbors(nodeids[8], depth=4, excludes=[nodes[4]])
    assert len(result) == 8
    assert nodes[4] not in result

    # get_neighbors_hc l5 in the middle expect: 25, 45, 65 and 85
    result = g.get_neighbors_hc(source=nodeids[4])
    exp_ids = [nodeids[1], nodeids[3], nodeids[5], nodeids[7]]
    exp_scidds = [scid25 + '/1', scid45 + '/0', scid56 + '/1', scid58 + '/0']
    assert len(result) == len(exp_ids)
    for halfchan in result:
        assert str(halfchan.source.node_id) == nodeids[4]
        assert str(halfchan.destination.node_id) in exp_ids
        assert str(halfchan) in exp_scidds

    # same but other direction
    result = g.get_neighbors_hc(destination=nodeids[4])
    exp_ids = [nodeids[1], nodeids[3], nodeids[5], nodeids[7]]
    exp_scidds = [scid25 + '/0', scid45 + '/1', scid56 + '/0', scid58 + '/1']
    assert len(result) == len(exp_ids)
    for halfchan in result:
        assert str(halfchan.destination.node_id) == nodeids[4]
        assert str(halfchan.source.node_id) in exp_ids
        assert str(halfchan) in exp_scidds

    # get all channels which have l1 as destination
    result = g.get_neighbors_hc(destination=nodeids[0])
    exp_ids = [nodeids[1], nodeids[3]]
    exp_scidds = [scid12 + '/0', scid14 + '/1']
    assert len(result) == len(exp_ids)
    for halfchan in result:
        assert str(halfchan.destination.node_id) == nodeids[0]
        assert str(halfchan.source.node_id) in exp_ids
        assert str(halfchan) in exp_scidds

    # l5 as destination in the middle but depth=1, so the outer ring
    # epxect: 12, 14, 32, 36, 74, 78, 98, 96
    result = g.get_neighbors_hc(destination=nodeids[4], depth=1)
    exp_scidds = [scid12 + '/1', scid14 + '/0', scid23 + '/1', scid36 + '/1',
                  scid47 + '/0', scid69 + '/1', scid78 + '/0', scid89 + '/0']
    assert len(result) == len(exp_scidds)
    for halfchan in result:
        assert str(halfchan) in exp_scidds

    # same but other direction
    result = g.get_neighbors_hc(source=nodeids[4], depth=1)
    exp_scidds = [scid12 + '/0', scid14 + '/1', scid23 + '/0', scid36 + '/0',
                  scid47 + '/1', scid69 + '/0', scid78 + '/1', scid89 + '/1']
    assert len(result) == len(exp_scidds)
    for halfchan in result:
        assert str(halfchan) in exp_scidds

    # l9 as destination and depth=2 expect: 23 25 45 47
    result = g.get_neighbors_hc(destination=nodeids[8], depth=2)
    exp_scidds = [scid23 + '/0', scid25 + '/0', scid45 + '/1', scid47 + '/1']
    assert len(result) == len(exp_scidds)
    for halfchan in result:
        assert str(halfchan) in exp_scidds

    # l9 as destination depth=2 exclude=[l7] expect: 23 25 45
    result = g.get_neighbors_hc(destination=nodeids[8], depth=2, excludes=[nodes[6]])
    exp_scidds = [scid23 + '/0', scid25 + '/0', scid45 + '/1']
    assert len(result) == len(exp_scidds)
    for halfchan in result:
        assert str(halfchan) in exp_scidds

    # same as above, but excludes halfchannels of l7 expect: 23 25 45
    hcs = [c.half_channels[0] for c in nodes[6].channels]
    hcs += [c.half_channels[1] for c in nodes[6].channels]
    result = g.get_neighbors_hc(destination=nodeids[8], depth=2, excludes=hcs)
    exp_scidds = [scid23 + '/0', scid25 + '/0', scid45 + '/1']
    assert len(result) == len(exp_scidds)
    for halfchan in result:
        assert str(halfchan) in exp_scidds

    # again, same as above, but excludes channels of l7 expect: 23 25 45
    chs = [c for c in nodes[6].channels]
    result = g.get_neighbors_hc(destination=nodeids[8], depth=2, excludes=chs)
    exp_scidds = [scid23 + '/0', scid25 + '/0', scid45 + '/1']
    assert len(result) == len(exp_scidds)
    for halfchan in result:
        assert str(halfchan) in exp_scidds

    # l9 as destination and depth=3 expect: 12 14
    result = g.get_neighbors_hc(destination=nodeids[8], depth=3)
    exp_scidds = [scid12 + '/1', scid14 + '/0']
    assert len(result) == len(exp_scidds)
    for halfchan in result:
        assert str(halfchan) in exp_scidds

    # l9 as destination and depth>=4 expect: empty set
    for d in range(4, 6):
        result = g.get_neighbors_hc(destination=nodeids[8], depth=d)
        assert len(result) == 0



from pyln.client.version import NodeVersion


def test_create_version():
    # These are the strings returned by `lightningd --version`
    _ = NodeVersion("v24.11-232-g5a76c7a")
    _ = NodeVersion("v24.11-225-gda793e6-modded")
    _ = NodeVersion("v24.11")
    _ = NodeVersion("vd6fa78c")


def test_equality_classes_in_node_versions():
    assert NodeVersion("v24.02") == NodeVersion("v24.02")
    assert NodeVersion("v24.02") == NodeVersion("24.02")
    assert NodeVersion("v24.02-225") == NodeVersion("v24.02")

    assert NodeVersion("v24.02") == NodeVersion("v24.02rc1")
    assert NodeVersion("v24.11-217-g77989b1-modded") == NodeVersion("v24.11")
    assert NodeVersion("vd6fa78c") == NodeVersion("vabcdefg")


def test_inequality_of_node_versions():
    assert not NodeVersion("v24.02.1") > NodeVersion("v24.02.1")
    assert NodeVersion("v24.02.1") > NodeVersion("v24.02")
    assert NodeVersion("v24.02.1") > NodeVersion("v24.02rc1")
    assert NodeVersion("v24.02.1") > NodeVersion("v23.05")
    assert NodeVersion("v24.05") > NodeVersion("v24.02")
    assert NodeVersion("vd6fa78c") > NodeVersion("v26.02")

    assert NodeVersion("v24.02.1") >= NodeVersion("v24.02.1")
    assert NodeVersion("v24.02.1") >= NodeVersion("v24.02")
    assert NodeVersion("v24.02.1") >= NodeVersion("v24.02rc1")
    assert NodeVersion("v24.02.1") >= NodeVersion("v23.05")
    assert NodeVersion("v24.05") >= NodeVersion("v24.02")
    assert NodeVersion("vd6fa78c") >= NodeVersion("v26.02")

    assert NodeVersion("v24.02.1") <= NodeVersion("v24.02.1")
    assert not NodeVersion("v24.02.1") <= NodeVersion("v24.02")
    assert not NodeVersion("v24.02.1") <= NodeVersion("v24.02rc1")
    assert not NodeVersion("v24.02.1") <= NodeVersion("v23.05")
    assert not NodeVersion("v24.05") <= NodeVersion("v24.02")
    assert not NodeVersion("vd6fa78c") <= NodeVersion("v26.02")

    assert not NodeVersion("v24.02.1") < NodeVersion("v24.02.1")
    assert not NodeVersion("v24.02.1") < NodeVersion("v24.02")
    assert not NodeVersion("v24.02.1") < NodeVersion("v24.02rc1")
    assert not NodeVersion("v24.02.1") < NodeVersion("v23.05")
    assert not NodeVersion("v24.05") < NodeVersion("v24.02")
    assert not NodeVersion("vd6fa78c") < NodeVersion("v26.02")
